-------------------- INIT --------------------------------

-- Some "constants" (not really)
local DEFAULT_POSX = -770
local DEFAULT_POSY = 410
local DEFAULT_ALPHA = 0.8
local DEFAULT_SIZE = 10
local DEFAULT_WIDTH = 1000
local DEFAULT_SHORTEN = false
local MIN_SIZE = 8
local MAX_SIZE = 28
local NEW_VERSION = 105

local ENUM_POSX = 1
local ENUM_POSY = 2
local ENUM_SIZE = 3
local ENUM_ALPHA = 4
local ENUM_VERSION = 5
local ENUM_SHORTEN = 6

local global_loop_interval = 0.1

-- id/intellect
local power_torrent = {74241, 500}
local volcanic_power = {79476, 1200}
local lightweave = {75170, 580}
local volcanic_destruction = {89091, 1600}
local synapse_springs = {96230, 480}

local blessing_of_kings = 79063
local mark_of_the_wild = 79061
		
local intellect_multiplier = 1

-- Examples like the following are probably redundant and hard to maintain:
--local well_fed = 87547
--local flask_of_the_draconic_mind = 92730

-- Create the main frame
local main_frame = CreateFrame("Frame", "AO_main_frame", UIParent); -- this had to be with parent on UIParent to not show during cinematics, otherwise, event checking spam would be needed.

-- Register Events
-- do stuff only when these events occur:
main_frame:RegisterEvent("ADDON_LOADED"); -- proper time to deal with saved variables
main_frame:RegisterEvent("PLAYER_ENTERING_WORLD"); 
main_frame:RegisterEvent("PLAYER_TALENT_UPDATE"); -- time to check if this is an Arcane Mage
main_frame:RegisterEvent("UNIT_POWER"); -- when mana changes
main_frame:RegisterEvent("UNIT_MAXPOWER") -- max mana changed
main_frame:RegisterEvent("UNIT_AURA") -- a buf drops off or is gained 
-- main_frame:RegisterEvent("MASTERY_UPDATE"); -- mastery changes
-- main_frame:RegisterEvent("PLAYER_DAMAGE_DONE_MODS"); -- spell bonus changed
main_frame:RegisterEvent("UNIT_STATS"); -- Primary stat (Intellect) changed; this is almost all of the time redundant due to UNIT_MAXPOWER but let's be paranoid.

-- Create a messaging frame
local msg_frame = CreateFrame("MessageFrame", "Msg_Frame", main_frame);
msg_frame:SetFading(false); -- do not fade
msg_frame:SetInsertMode("BOTTOM"); -- start from bottom
msg_frame:SetJustifyH("RIGHT"); -- justify right to arrange numbers in a readable fashion

-- Some booleans:
local arcane = false; -- is it an arcane mage? 
local spammed = false; -- don't spam the loading mes

-- This is called initially, between spec changes, and after a frame clear: 
local function init_saved_vars()
	
	if AOSettings == nil then
		AOSettings = {}
	end


	-- for future version upgrade fixes
	if tonumber(AOSettings[ENUM_VERSION]) == nil then 
	end
	AOSettings[ENUM_VERSION] = NEW_VERSION

	-- PosX/Y
	if tonumber(AOSettings[ENUM_POSX]) == nil or tonumber(AOSettings[ENUM_POSY]) == nil then -- no config file or empty strings
		AOSettings[ENUM_POSX] = DEFAULT_POSX; AOSettings[ENUM_POSY] = DEFAULT_POSY; -- set to defaults; we do not directly use them because we want them saved
	end
	msg_frame:SetPoint("BOTTOMRIGHT",nil, "BOTTOMRIGHT", AOSettings[ENUM_POSX], AOSettings[ENUM_POSY]);
		
	-- Alpha
	if tonumber(AOSettings[ENUM_ALPHA]) == nil then
		AOSettings[ENUM_ALPHA] = DEFAULT_ALPHA;
	end
	msg_frame:SetAlpha(AOSettings[ENUM_ALPHA]); 

	-- Size based on font size
	if tonumber(AOSettings[ENUM_SIZE]) == nil then
		AOSettings[ENUM_SIZE] = DEFAULT_SIZE;
	end

	msg_frame:SetFont("Fonts\\FRIZQT__.TTF", AOSettings[ENUM_SIZE], "OUTLINE");
	msg_frame:SetWidth(DEFAULT_WIDTH); -- sufficiently big and constant 
	msg_frame:SetHeight(AOSettings[ENUM_SIZE]*3 - 1); -- '*3' appears inadequade, more than 4 may show 4 lines. Just 4 may in rare cases shows 4 lines but not certain. It appears there are rounding system quirks.
	
end
--

-- helper functions: 
-- round a number, from http://lua-users.org/wiki/FormattingNumbers
local function round(val, decimal)
	if (decimal) then
		return math.floor( (val * 10^decimal) + 0.5) / (10^decimal)
	else
		return math.floor(val+0.5)
	end
end

-- Interface GUI: 

local Interface = {};

Interface.panel = CreateFrame( "Frame", "AOInterfacePanel", UIParent ); 

Interface.panel.name = "Arcane Overflow";

-- Add the panel to the Interface Options
InterfaceOptions_AddCategory(Interface.panel);

-- posx/y sliders

local slider_posx = CreateFrame("Slider","AOslider_posx",AOInterfacePanel,"OptionsSliderTemplate") -- name is required for later use of getglobal
slider_posx:SetPoint("TOPLEFT",20,-30)
getglobal(slider_posx:GetName() .. 'Low'):SetText('Left')
getglobal(slider_posx:GetName() .. 'High'):SetText('Right')
getglobal(slider_posx:GetName() .. 'Text'):SetText('Horizontal Position') 
slider_posx:SetWidth(360)
slider_posx:SetMinMaxValues(-1280,0)
slider_posx:SetValueStep(10)

local slider_posy = CreateFrame("Slider","AOslider_posy",AOInterfacePanel,"OptionsSliderTemplate") -- name is required for later use of getglobal
slider_posy:SetPoint("TOPLEFT",20,-80)
getglobal(slider_posy:GetName() .. 'Low'):SetText('Bottom')
getglobal(slider_posy:GetName() .. 'High'):SetText('Top')
getglobal(slider_posy:GetName() .. 'Text'):SetText('Vertical Position') 
slider_posy:SetWidth(360)
slider_posy:SetMinMaxValues(0,730)
slider_posy:SetValueStep(10)

-- size slider
local slider_size = CreateFrame("Slider","AOslider_size",AOInterfacePanel,"OptionsSliderTemplate") -- name is required for later use of getglobal
slider_size:SetPoint("TOPLEFT",20,-130)
getglobal(slider_size:GetName() .. 'Low'):SetText('Small')
getglobal(slider_size:GetName() .. 'High'):SetText('Big')
getglobal(slider_size:GetName() .. 'Text'):SetText('Size') 
slider_size:SetWidth(360)
slider_size:SetMinMaxValues(MIN_SIZE,MAX_SIZE)
slider_size:SetValueStep(1)

-- alpha slider
local slider_alpha = CreateFrame("Slider","AOslider_alpha",AOInterfacePanel,"OptionsSliderTemplate") 
slider_alpha:SetPoint("TOPLEFT",20,-180)
getglobal(slider_alpha:GetName() .. 'Low'):SetText('Transparent')
getglobal(slider_alpha:GetName() .. 'High'):SetText('Solid')
getglobal(slider_alpha:GetName() .. 'Text'):SetText('Alpha')
slider_alpha:SetWidth(360)
slider_alpha:SetMinMaxValues(0,1)
slider_alpha:SetValueStep(0.05)

-- check box
local check_box = CreateFrame("CheckButton", "AOcheck_box", AOInterfacePanel, "ChatConfigCheckButtonTemplate ")
check_box:SetPoint("TOPLEFT",270,-210)
AOcheck_boxText:SetText("Short Text")

-- reset button
local button_reset = CreateFrame("Button","AObutton_reset",AOInterfacePanel,"UIPanelButtonTemplate")
button_reset:SetPoint("TOPLEFT",270,-240)
button_reset:SetWidth(120)
button_reset:SetHeight(20)
button_reset:SetText("Reset all options")


--------------------- END OF INIT ------------------------

local arcane_overflow = 0
local soonest_expiration = {}
soonest_expiration[1] = 0
soonest_expiration[2] = 0

-- Set up handling of events
local function Handle_Events(self, event, ...)
	
	if event == "ADDON_LOADED" or event == "PLAYER_TALENT_UPDATE" then -- for some reason PLAYER_TALENT_UPDATE occurs even after a /reload; it's probable ADDON_LOADED is unneeded here, though the extra load  inflicted is unimportant if it is.
		local playerClass, englishClass = UnitClass("player");
		if englishClass == "MAGE" then
		
			local _, _, _, _, arcane_points = GetTalentTabInfo(1);
        		local _, _, _, _, fire_points = GetTalentTabInfo(2);   
        		local _, _, _, _, frost_points = GetTalentTabInfo(3);

			if arcane_points > fire_points and arcane_points > frost_points then
				if spammed == false then
					print("|cFF0098f8Arcane Mage detected; Spell Power Adept is shown.");
					init_saved_vars(); -- do initialization; this is also required after a clear;
					spammed = true;   -- do it only once per run since even a short-term anti-spam method with time() was unreliable, perhaps due to game bugs or quirks; combustionhelper though does it properly somehow. 
				end
				arcane = true;
			else
				arcane = false;
				msg_frame:Clear(); -- clear the frame
			end
		else -- not a mage at all
			arcane = false;
			msg_frame:Clear();
		end

	end

	if arcane then -- our common routine:

		-- Current mana:
		local mana = UnitPower("player");
		-- Current Max Mana:
		local current_max_mana = UnitPowerMax("player");
		-- extra mana from other sources, calculated later
		local extra_mana = 0

		--[[ mage ward and mana shield tests
		local test = {543, 500}
		local test2 = {1463,300}
		--]]

		-- see if we have kings
		for i = 1, 40 do
			local name, _, _, count, _, _, expirationTime, _, _, _, spellId = UnitAura("player", i)
			if (spellId == blessing_of_kings or spellId  == mark_of_the_wild) then
				intellect_multiplier = 1.05;
			end
		end
		
		local found_expiration_id = 1
		for i = 1, 40 do
			
			local name, _, _, count, _, _, expirationTime, _, _, _, spellId = UnitAura("player", i)

		if (spellId == power_torrent[1]) then
				-- for some weird reason, it needs the addition of the previous here as well:
				extra_mana = extra_mana + power_torrent[2] * 15 * intellect_multiplier
				-- take this expiration as the soonest if it's later in time
				if (soonest_expiration[2] == 0 or soonest_expiration[2] < expirationTime) then
					soonest_expiration[1] = power_torrent[1]
					soonest_expiration[2] = expirationTime
				end
			
				-- see if an expiration time currently running was dropped
				if soonest_expiration [1] ==  spellId then
					found_expiration_id = 0
				end
			end
			-- copied procedure
			if (spellId == volcanic_power[1]) then
				-- add to the previous:
				extra_mana = extra_mana + volcanic_power[2] * 15 * intellect_multiplier
				if (soonest_expiration[2] == 0 or soonest_expiration[2] < expirationTime) then
					soonest_expiration[1] = volcanic_power[1]
					soonest_expiration[2] = expirationTime
				end
			
				if soonest_expiration [1] ==  spellId then
					found_expiration_id = 0
				end
			end
			-- similarly
			if (spellId == lightweave[1]) then
				extra_mana = extra_mana + lightweave[2] * 15 * intellect_multiplier
				if (soonest_expiration[2] == 0 or soonest_expiration[2] < expirationTime) then
					soonest_expiration[1] = lightweave[1]
					soonest_expiration[2] = expirationTime
				end
			
				if soonest_expiration [1] ==  spellId then
					found_expiration_id = 0
				end
			end
			if (spellId == volcanic_destruction[1]) then
				extra_mana = extra_mana + volcanic_destruction[2] * 15 * intellect_multiplier
				if (soonest_expiration[2] == 0 or soonest_expiration[2] < expirationTime) then
					soonest_expiration[1] = volcanic_destruction[1]
					soonest_expiration[2] = expirationTime
				end
			
				if soonest_expiration [1] ==  spellId then
					found_expiration_id = 0
				end
			end
			if (spellId == synapse_springs[1]) then
				extra_mana = extra_mana + synapse_springs[2] * 15 * intellect_multiplier
				if (soonest_expiration[2] == 0 or soonest_expiration[2] < expirationTime) then
					soonest_expiration[1] = synapse_springs[1]
					soonest_expiration[2] = expirationTime
				end
			
				if soonest_expiration [1] ==  spellId then
					found_expiration_id = 0
				end
			end

			--[[
			if (spellId == test[1]) then
				-- for some weird reason, it needs the addition of the previous here as well:
				extra_mana = extra_mana + test[2] * 15 * intellect_multiplier
				-- take this expiration as the soonest if it's later in time
				if (soonest_expiration[2] == 0 or soonest_expiration[2] < expirationTime) then
					soonest_expiration[1] = test[1]
					soonest_expiration[2] = expirationTime
				end
			
				-- see if an expiration time currently running was dropped
				if soonest_expiration [1] ==  spellId then
					found_expiration_id = 0
				end
			end
			-- copied procedure
			if (spellId == test2[1]) then
				-- add to the previous:
				extra_mana = extra_mana + test2[2] * 15 * intellect_multiplier
				if (soonest_expiration[2] == 0 or soonest_expiration[2] < expirationTime) then
					soonest_expiration[1] = test2[1]
					soonest_expiration[2] = expirationTime
				end
			
				if soonest_expiration [1] ==  spellId then
					found_expiration_id = 0
				end
			end
			--]]
		
		end
		-- Remove an expiration time if its buf was dropped
		if found_expiration_id ~= 0 then
			soonest_expiration[1] = 0
			soonest_expiration[2] = 0
		end
		
		-- Arcane Overflow:
		arcane_overflow = mana + extra_mana - current_max_mana 
		if arcane_overflow < 0 then arcane_overflow = 0 end

		-- Also remove the expiration time if there is simply no overflow
		if (arcane_overflow == 0) then
			soonest_expiration[1] = 0
			soonest_expiration[2] = 0
		end

	end
end

-- Mainly for printing since expiration time updating does not usually coincide with events
local function loop()

	local AO_text
	local ETR_text

	if AOSettings[ENUM_SHORTEN] then
		AO_text = "|cFF0098f8O: |r"
		ETR_text = "|cFF0098f8T: |r"
	else
		AO_text = "|cFF0098f8Overflow: |r"
		ETR_text = "|cFF0098f8Remaining: |r"
	end

	msg_frame:AddMessage(AO_text .. arcane_overflow);
	
	local soonest_timeout = 0
	if soonest_expiration[2] then
		soonest_timeout = soonest_expiration[2]- GetTime()
	end
	if soonest_timeout < 0 then soonest_timeout = 0 end
	msg_frame:AddMessage(ETR_text .. round(soonest_timeout,1) .. "s");

end

-- OnUpdate handling the loop every global_loop_interval
local total = 0
local function onUpdate(self,elapsed)
	total = total + elapsed
	if total >= global_loop_interval then
		loop()
		total = 0
	end
end
 
local f = CreateFrame("frame")
main_frame:SetScript("OnUpdate", onUpdate)


-- Handle events
main_frame:SetScript("OnEvent", Handle_Events);

-- Interface GUI:

-- When the panel is shown refresh the values and warn in case it's disabled boxes
local function  InterfaceRefreshValues()
	if arcane == false then end
	slider_posx:SetValue(AOSettings[ENUM_POSX])
	slider_posy:SetValue(AOSettings[ENUM_POSY])
	slider_alpha:SetValue(AOSettings[ENUM_ALPHA])
	slider_size:SetValue(AOSettings[ENUM_SIZE])
	check_box:SetChecked(AOSettings[ENUM_SHORTEN])

end
Interface.panel:SetScript("OnShow", InterfaceRefreshValues);

-- Reset Button:
local function Reset()
	if arcane == false then return; end 
	AOSettings[ENUM_POSX] = DEFAULT_POSX; AOSettings[ENUM_POSY] = DEFAULT_POSY; -- defaults; we do not directly use them because we want them saved
	msg_frame:SetPoint("BOTTOMRIGHT",nil, "BOTTOMRIGHT", AOSettings[ENUM_POSX], AOSettings[ENUM_POSY]);

	AOSettings[ENUM_ALPHA] = DEFAULT_ALPHA; 
	msg_frame:SetAlpha(AOSettings[ENUM_ALPHA]); 

	AOSettings[ENUM_SIZE] = DEFAULT_SIZE;
	msg_frame:SetFont("Fonts\\FRIZQT__.TTF", AOSettings[ENUM_SIZE], "OUTLINE");
	msg_frame:SetWidth(DEFAULT_WIDTH); 
	msg_frame:SetHeight(AOSettings[ENUM_SIZE]*3 - 1);

	AOSettings[ENUM_SHORTEN] = DEFAULT_SHORTEN;
	check_box:SetChecked(AOSettings[ENUM_SHORTEN])

	-- Refresh values on frame
	InterfaceRefreshValues()

	print("|cFF0098f8Arcane Overflow settings reset");
end
button_reset:SetScript("PostClick", Reset)

-- PosX/Y Sliders
local function slider_posx_function()
	if arcane == false then return; end
	AOSettings[ENUM_POSX] = slider_posx:GetValue()
	if tonumber(AOSettings[ENUM_POSX]) == nil then AOSettings[ENUM_POSX] = DEFAULT_POSX; InterfaceRefreshValues(); end
	slider_posx.tooltipText = AOSettings[ENUM_POSX]
	msg_frame:SetPoint("BOTTOMRIGHT",nil, "BOTTOMRIGHT", AOSettings[ENUM_POSX], AOSettings[ENUM_POSY]);
end
slider_posx:SetScript("OnValueChanged", slider_posx_function)

local function slider_posy_function()
	if arcane == false then return; end
	AOSettings[ENUM_POSY] = slider_posy:GetValue()
	if tonumber(AOSettings[ENUM_POSY]) == nil then AOSettings[ENUM_POSY] = DEFAULT_POSY; InterfaceRefreshValues(); end
	slider_posy.tooltipText = AOSettings[ENUM_POSY]
	msg_frame:SetPoint("BOTTOMRIGHT",nil, "BOTTOMRIGHT", AOSettings[ENUM_POSX], AOSettings[ENUM_POSY]);
end
slider_posy:SetScript("OnValueChanged", slider_posy_function)


-- alpha slider
local function slider_alpha_function()
	if arcane == false then return; end
	AOSettings[ENUM_ALPHA] = slider_alpha:GetValue()
	if tonumber(AOSettings[ENUM_ALPHA]) == nil then AOSettings[ENUM_ALPHA] = DEFAULT_ALPHA; InterfaceRefreshValues(); end
	slider_alpha.tooltipText = round(AOSettings[ENUM_ALPHA],2) -- who knows why, it may not be round, maybe a system quirk
	msg_frame:SetAlpha(AOSettings[ENUM_ALPHA])
end
slider_alpha:SetScript("OnValueChanged", slider_alpha_function)

-- size slider; playing with this may produce visual quirks that go away with a /reload. Perhaps a system quirk.
local function slider_size_function()
	if arcane == false then return; end
	AOSettings[ENUM_SIZE] = slider_size:GetValue()
	if tonumber(AOSettings[ENUM_SIZE]) == nil then AOSettings[ENUM_SIZE] = DEFAULT_ALPHA; InterfaceRefreshValues(); end
	slider_size.tooltipText = AOSettings[ENUM_SIZE]
	msg_frame:SetFont("Fonts\\FRIZQT__.TTF", AOSettings[ENUM_SIZE], "OUTLINE");
	msg_frame:SetWidth(DEFAULT_WIDTH); 
	msg_frame:SetHeight(AOSettings[ENUM_SIZE]*3 - 1);
end
slider_size:SetScript("OnValueChanged", slider_size_function)

-- check box function
local function check_box_function()
	if arcane == false then return; end
	AOSettings[ENUM_SHORTEN] = check_box:GetChecked()
end
check_box:SetScript("OnClick", check_box_function)

-- Slash command: (there were others but were dropped to ease maintenance for a feature seldom needed)

SLASH_AO1, SLASH_AO2 = '/ao', '/ArcaneOverflow';
function SlashCmdList.AO(msg, editbox)
	InterfaceOptionsFrame_OpenToCategory(Interface.panel)
end
